

set -ex



test -L $PREFIX/lib/libcublas.so
test -L $PREFIX/lib/libcublasLt.so
test -L $PREFIX/lib/libnvblas.so
test -f $PREFIX/lib/pkgconfig/cublas*.pc
test -f $PREFIX/targets/sbsa-linux/include/cublas.h
test -f $PREFIX/targets/sbsa-linux/include/cublasLt.h
test -f $PREFIX/targets/sbsa-linux/include/nvblas.h
test -L $PREFIX/targets/sbsa-linux/lib/libcublas.so
test -L $PREFIX/targets/sbsa-linux/lib/libcublasLt.so
test -L $PREFIX/targets/sbsa-linux/lib/libnvblas.so
test -f $PREFIX/targets/sbsa-linux/lib/stubs/libcublas.so
test -f $PREFIX/targets/sbsa-linux/lib/stubs/libcublasLt.so
exit 0
